#pragma once

// Used for returning errors from sounddevice functions.
enum SoundDeviceError {
    SOUNDDEVICE_ERROR_ERR = -1,
    SOUNDDEVICE_ERROR_OK = 0,
    SOUNDDEVICE_ERROR_DUPLICATE_OUTPUT_CHANNEL,
    SOUNDDEVICE_ERROR_EXCESSIVE_OUTPUT_CHANNEL,
    SOUNDDEVICE_ERROR_EXCESSIVE_INPUT_CHANNEL,
    SOUNDDEVICE_ERROR_DEVICE_COUNT
};
