// This is just a dummy file meant to be shoved in the build list so that moc
// runs on the class and linker errors are avoided.

#include "util/sleepableqthread.h"

#include "moc_sleepableqthread.cpp"
